from typing import Iterable, Sequence, Tuple, TypeVar

T = TypeVar("T")


def grouped(items: Sequence[T], n: int = 2) -> Iterable[Tuple[T, ...]]:
    """s -> [(s0, s1, s2,...sn-1), (sn, sn+1 , sn+2,...s2n-1), ...]
    list(grouped([1, 2], 2)) == [(1, 2)]
    list(grouped([1, 2, 3, 4], 2)) == [(1, 2), (3, 4)]
    """
    if len(items) % n != 0:
        raise ValueError("items must be equally divisible by n")
    return zip(*[iter(items)] * n)



import inspect
import typing

from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from pydantic.schema import get_flat_models_from_field
from starlette.datastructures import FormData, UploadFile
from starlette.requests import HTTPConnection, Request

import xpresso.openapi.models as openapi_models
from xpresso._utils.pydantic_utils import is_sequence_like, model_field_from_param
from xpresso._utils.schemas import openapi_schema_from_pydantic_field
from xpresso._utils.typing import get_args, get_type_hints
from xpresso.binders._binders.formencoded_parsing import Extractor as FormDataExtractor
from xpresso.binders._binders.formencoded_parsing import (
    InvalidSerialization,
    get_extractor,
)
from xpresso.binders._binders.media_type_validator import MediaTypeValidator
from xpresso.binders._binders.pydantic_validators import validate_body_field
from xpresso.binders.api import (
    ModelNameMap,
    OpenAPIMetadata,
    SupportsExtractor,
    SupportsOpenAPI,
)
from xpresso.exceptions import RequestValidationError
from xpresso.openapi._utils import parse_examples
from xpresso.typing import Some


class FormFieldExtractor(typing.NamedTuple):
    style: str
    explode: bool
    field_name: str
    extractor: FormDataExtractor

    async def extract(self, form: FormData) -> typing.Optional[Some]:
        params = [(k, v) for k, v in form.multi_items() if isinstance(v, str)]  # type: ignore
        try:
            return self.extractor(name=self.field_name, params=params)
        except InvalidSerialization as e:
            raise RequestValidationError(
                [
                    ErrorWrapper(
                        exc=TypeError("Data is not a valid URL encoded form"),
                        loc=tuple(("body", self.field_name)),
                    )
                ]
            ) from e


class FormFieldExtractorMarker(typing.NamedTuple):
    alias: typing.Optional[str]
    style: str
    explode: bool

    def register_parameter(self, param: inspect.Parameter) -> FormFieldExtractor:
        field = model_field_from_param(param, alias=self.alias)
        extractor = get_extractor(style=self.style, explode=self.explode, field=field)
        return FormFieldExtractor(
            style=self.style,
            explode=self.explode,
            field_name=field.name,
            extractor=extractor,
        )


class FormFieldOpenAPIMetadata(typing.NamedTuple):
    schema: openapi_models.Schema
    encoding: openapi_models.Encoding


class FormFieldOpenAPI(typing.NamedTuple):
    field_name: str
    style: str
    explode: bool
    field: ModelField

    def get_models(self) -> typing.List[type]:
        return list(get_flat_models_from_field(self.field, set()))

    def get_field_openapi(
        self, model_name_map: ModelNameMap, schemas: typing.Dict[str, typing.Any]
    ) -> FormFieldOpenAPIMetadata:
        field_schema = openapi_schema_from_pydantic_field(
            self.field, model_name_map, schemas
        )
        return FormFieldOpenAPIMetadata(
            schema=field_schema,
            encoding=openapi_models.Encoding(
                contentType=None,
                style=self.style,
                explode=self.explode,
            ),
        )


class FormFieldOpenAPIMarker(typing.NamedTuple):
    alias: typing.Optional[str]
    style: str
    explode: bool

    def register_parameter(self, param: inspect.Parameter) -> FormFieldOpenAPI:
        field = model_field_from_param(param, alias=self.alias)
        return FormFieldOpenAPI(
            field_name=field.name,
            style=self.style,
            explode=self.explode,
            field=field,
        )


def ensure_field_is_a_file(
    field: typing.Union[str, UploadFile], field_name: str
) -> UploadFile:
    if isinstance(field, str):
        raise RequestValidationError(
            [
                ErrorWrapper(
                    exc=TypeError("Expected a file, got a string"),
                    loc=("body", field_name),
                )
            ]
        )
    return field


class FormFileExtractor(typing.NamedTuple):
    consumer: typing.Callable[[UploadFile], typing.Awaitable[typing.Any]]
    repeated: bool
    field_name: str

    async def extract(
        self,
        form: FormData,
    ) -> typing.Optional[Some]:
        if self.repeated:
            files: "typing.List[typing.Union[bytes, UploadFile]]" = []
            for field_name, field_value in form.multi_items():
                if field_name == self.field_name:
                    file = ensure_field_is_a_file(field_value, field_name)
                    files.append(await self.consumer(file))
            return Some(files)
        if self.field_name not in form:
            return None
        file = ensure_field_is_a_file(form[self.field_name], self.field_name)
        return Some(await self.consumer(file))


class FormFileExtractorMarker(typing.NamedTuple):
    alias: typing.Optional[str]

    def register_parameter(self, param: inspect.Parameter) -> FormFileExtractor:
        field = model_field_from_param(param, arbitrary_types_allowed=True)
        repeated = is_sequence_like(field)
        if field.type_ is bytes:

            async def read_uploadfile_to_bytes(file: UploadFile) -> bytes:
                return await file.read()  # type: ignore[return-value]

            return FormFileExtractor(
                read_uploadfile_to_bytes,
                field_name=self.alias or param.name,
                repeated=repeated,
            )
        elif inspect.isclass(field.type_) and issubclass(field.type_, UploadFile):

            async def read_uploadfile_to_uploadfile(file: UploadFile) -> UploadFile:
                return file

            return FormFileExtractor(
                read_uploadfile_to_uploadfile,
                field_name=self.alias or param.name,
                repeated=repeated,
            )
        else:
            raise TypeError(f"Unknown file type {field.type_.__name__}")


class FormFileOpenAPI(typing.NamedTuple):
    media_type: typing.Optional[str]
    format: str
    nullable: bool
    repeated: bool
    field_name: str

    def get_models(self) -> typing.List[type]:
        return []

    def get_field_openapi(
        self, model_name_map: ModelNameMap, schemas: typing.Dict[str, typing.Any]
    ) -> FormFieldOpenAPIMetadata:
        schema = openapi_models.Schema(
            type="string", format=self.format, nullable=self.nullable or None
        )
        if self.repeated:
            schema = openapi_models.Schema(type="array", items=schema)
        return FormFieldOpenAPIMetadata(
            schema=schema,
            encoding=openapi_models.Encoding(contentType=self.media_type),
        )


class FormFileOpenAPIMarker(typing.NamedTuple):
    media_type: typing.Optional[str]
    format: str
    alias: typing.Optional[str]

    def register_parameter(self, param: inspect.Parameter) -> FormFileOpenAPI:
        field = model_field_from_param(param, arbitrary_types_allowed=True)
        return FormFileOpenAPI(
            field_name=self.alias or param.name,
            media_type=self.media_type,
            format=self.format,
            nullable=field.allow_none,
            repeated=is_sequence_like(field),
        )


class FormFieldMarker(typing.NamedTuple):
    extractor_marker: typing.Union[FormFieldExtractorMarker, FormFileExtractorMarker]
    openapi_marker: typing.Union[FormFieldOpenAPIMarker, FormFileOpenAPIMarker]


class Extractor(typing.NamedTuple):
    field: ModelField
    field_extractors: typing.Mapping[
        str, typing.Union[FormFileExtractor, FormFieldExtractor]
    ]
    media_type_validator: MediaTypeValidator

    def __hash__(self) -> int:
        return hash("form")

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, Extractor)

    async def extract(self, connection: HTTPConnection) -> typing.Any:
        assert isinstance(connection, Request)
        content_type = connection.headers.get("content-type", None)
        if (
            content_type is None
            and connection.headers.get("content-length", "0") == "0"
        ):
            return validate_body_field(None, field=self.field, loc=("body",))
        self.media_type_validator.validate(content_type)
        form = await connection.form()
        res: typing.Dict[str, typing.Any] = {}
        for param_name, extractor in self.field_extractors.items():
            extracted = await extractor.extract(form)
            if isinstance(extracted, Some):
                res[param_name] = extracted.value
        return validate_body_field(
            Some(res),
            field=self.field,
            loc=("body",),
        )


class ExtractorMarker(typing.NamedTuple):
    media_type: str

    def register_parameter(self, param: inspect.Parameter) -> SupportsExtractor:
        form_data_field = model_field_from_param(param)
        field_extractors: typing.Dict[
            str, typing.Union[FormFileExtractor, FormFieldExtractor]
        ] = {}
        # use pydantic to get rid of outer annotated, optional, etc.
        model = form_data_field.type_
        # workaround https://github.com/samuelcolvin/pydantic/pull/3413
        # by using get_type_hints
        type_hints = get_type_hints(model, include_extras=True)
        for field_param in inspect.signature(model).parameters.values():
            field_param = field_param.replace(annotation=type_hints[field_param.name])
            for m in get_args(field_param.annotation):
                if isinstance(m, (FormFieldMarker)):
                    field_extractor = m.extractor_marker.register_parameter(field_param)
                    break
            else:
                field_extractor = FormFieldExtractorMarker(
                    alias=None, style="form", explode=False
                ).register_parameter(field_param)
            field_extractors[field_param.name] = field_extractor
        return Extractor(
            media_type_validator=MediaTypeValidator(self.media_type),
            field_extractors=field_extractors,
            field=form_data_field,
        )


class OpenAPI(typing.NamedTuple):
    field_openapi_providers: typing.Mapping[
        str, typing.Union[FormFieldOpenAPI, FormFileOpenAPI]
    ]
    required_fields: typing.List[str]
    description: typing.Optional[str]
    examples: typing.Optional[openapi_models.Examples]
    media_type: str
    required: bool
    nullable: bool
    include_in_schema: bool

    def get_models(self) -> typing.List[type]:
        return [
            model
            for provider in self.field_openapi_providers.values()
            for model in provider.get_models()
        ]

    def get_openapi(
        self,
        model_name_map: ModelNameMap,
    ) -> OpenAPIMetadata:
        if not self.include_in_schema:
            return OpenAPIMetadata()
        schemas: typing.Dict[str, typing.Any] = {}
        providers_openapis = {
            field_name: field_openapi.get_field_openapi(
                model_name_map=model_name_map,
                schemas=schemas,
            )
            for field_name, field_openapi in self.field_openapi_providers.items()
        }
        properties = {
            field_name: openapi_meta.schema
            for field_name, openapi_meta in providers_openapis.items()
        }
        encodings = {
            field_name: openapi_meta.encoding
            for field_name, openapi_meta in providers_openapis.items()
        }
        schema = openapi_models.Schema(
            type="object",
            properties=properties,
            required=self.required_fields or None,
            nullable=self.nullable or None,
        )
        media_type = openapi_models.MediaType(
            schema=schema,  # type: ignore
            examples=self.examples,
            encoding=encodings or None,
        )
        return OpenAPIMetadata(
            body=openapi_models.RequestBody(
                content={self.media_type: media_type},
                required=self.required,
            ),
            schemas=schemas,
        )


class OpenAPIMarker(typing.NamedTuple):
    description: typing.Optional[str]
    examples: typing.Optional[
        typing.Dict[str, typing.Union[openapi_models.Example, typing.Any]]
    ]
    media_type: str
    include_in_schema: bool

    def register_parameter(self, param: inspect.Parameter) -> SupportsOpenAPI:
        form_data_field = model_field_from_param(param)
        required = form_data_field.required is not False
        field_openapi_providers: typing.Dict[
            str, typing.Union[FormFieldOpenAPI, FormFileOpenAPI]
        ] = {}
        required_fields: typing.List[str] = []
        # use pydantic to get rid of outer annotated, optional, etc.
        model = form_data_field.type_
        for field_param in inspect.signature(model).parameters.values():
            for m in get_args(field_param.annotation):
                if isinstance(m, FormFieldMarker):
                    field_openapi = m.openapi_marker.register_parameter(field_param)
                    break
            else:
                field_openapi = FormFieldOpenAPIMarker(
                    alias=None, style="form", explode=True
                ).register_parameter(field_param)
            field_name = field_openapi.field_name
            field_openapi_providers[field_name] = field_openapi
            if (
                model_field_from_param(
                    field_param, arbitrary_types_allowed=True
                ).required
                is not False
            ):
                required_fields.append(field_name)
        examples = parse_examples(self.examples) if self.examples else None
        return OpenAPI(
            field_openapi_providers=field_openapi_providers,
            required_fields=required_fields,
            description=self.description,
            examples=examples,
            media_type=self.media_type,
            required=required,
            nullable=form_data_field.allow_none,
            include_in_schema=self.include_in_schema,
        )



import inspect
import typing

import xpresso.openapi.models as openapi_models
from xpresso._utils.pydantic_utils import model_field_from_param
from xpresso._utils.typing import Annotated, get_args, get_origin
from xpresso.binders.api import (
    ModelNameMap,
    OpenAPIMetadata,
    SupportsExtractor,
    SupportsOpenAPI,
)
from xpresso.binders.dependants import Binder, BinderMarker
from xpresso.exceptions import HTTPException, RequestValidationError
from xpresso.requests import HTTPConnection, Request


def get_binders_from_union_annotation(param: inspect.Parameter) -> typing.List[Binder]:
    providers: typing.List[Binder] = []
    annotation = param.annotation
    origin = get_origin(annotation)
    assert origin is Annotated  # if we got called, this must be true
    annotation = next(iter(get_args(annotation)))
    origin = get_origin(annotation)
    if origin is not typing.Union:
        raise TypeError("Outermost type must be a Union")
    args = get_args(annotation)
    for arg in args:
        origin = get_origin(arg)
        assert origin is Annotated  # if we got called, this must be true
        marker = next((m for m in get_args(arg) if isinstance(m, BinderMarker)), None)
        if marker is None:
            raise TypeError(f"Type annotation is missing body marker: {arg}")
        providers.append(marker.register_parameter(param.replace(annotation=arg)))
    return providers


class BodyOpenAPI(typing.NamedTuple):
    description: typing.Optional[str]
    required: bool
    providers: typing.List[SupportsOpenAPI]

    def get_models(self) -> typing.List[type]:
        return [model for p in self.providers for model in p.get_models()]

    def get_openapi(self, model_name_map: ModelNameMap) -> OpenAPIMetadata:
        content: typing.Dict[str, openapi_models.MediaType] = {}
        schemas: typing.Dict[str, typing.Any] = {}
        for provider in self.providers:
            meta = provider.get_openapi(model_name_map)
            if meta.body is None:
                raise TypeError()
            for k, v in meta.body.content.items():
                if k in content:
                    raise ValueError(f"Duplicate bodies for media type {k}")
                content[k] = v
            schemas.update(meta.schemas or {})
        body = openapi_models.RequestBody(
            required=self.required, content=content, description=self.description
        )
        return OpenAPIMetadata(body=body, schemas=schemas)


class BodyOpenAPIMarker(typing.NamedTuple):
    description: typing.Optional[str]

    def register_parameter(self, param: inspect.Parameter) -> SupportsOpenAPI:
        field = model_field_from_param(param)
        return BodyOpenAPI(
            providers=[b.openapi for b in get_binders_from_union_annotation(param)],
            required=field.required is not False,
            description=self.description,
        )


class Extractor(typing.NamedTuple):
    providers: typing.Tuple[SupportsExtractor, ...]

    async def extract(self, connection: HTTPConnection) -> typing.Any:
        assert isinstance(connection, Request)
        errors: "typing.List[typing.Union[HTTPException, RequestValidationError]]" = []
        for provider in self.providers:
            try:
                return await provider.extract(connection)
            except (HTTPException, RequestValidationError) as error:
                errors.append(error)
        # if any body accepted the request but didn't pass validation, return the error from that one
        val_err = next((e for e in errors if e.status_code == 422), None)
        if val_err is not None:
            raise val_err
        # otherwise, jsut raise the first error we found
        # this is somewhat arbitrary, but there really is no good way to "merge"
        # all of the errors without making it confusing to the client
        # and leaking implementation details
        raise next(iter(errors))


class ExtractorMarker(typing.NamedTuple):
    def register_parameter(self, param: inspect.Parameter) -> SupportsExtractor:
        return Extractor(
            tuple(b.extractor for b in get_binders_from_union_annotation(param))
        )



from starlette import status as status
from starlette.exceptions import HTTPException
from starlette.responses import Response

from xpresso.applications import App
from xpresso.bodies import (
    BodyUnion,
    File,
    Form,
    FormField,
    FormFile,
    FromBodyUnion,
    FromFile,
    FromFormData,
    FromFormField,
    FromFormFile,
    FromJson,
    FromMultipart,
    Json,
    Multipart,
)
from xpresso.datastructures import UploadFile
from xpresso.dependencies import Depends
from xpresso.exception_handlers import ExcHandler
from xpresso.parameters import (
    CookieParam,
    FromCookie,
    FromHeader,
    FromPath,
    FromQuery,
    HeaderParam,
    PathParam,
    QueryParam,
)
from xpresso.requests import Request
from xpresso.routing.operation import Operation
from xpresso.routing.pathitem import Path
from xpresso.routing.router import Router
from xpresso.routing.websockets import WebSocketRoute
from xpresso.websockets import WebSocket

__all__ = (
    "ExcHandler",
    "Operation",
    "Path",
    "QueryParam",
    "HeaderParam",
    "CookieParam",
    "PathParam",
    "Json",
    "FormFile",
    "FromFormFile",
    "Form",
    "File",
    "Multipart",
    "Depends",
    "App",
    "Router",
    "UploadFile",
    "FromBodyUnion",
    "BodyUnion",
    "FromCookie",
    "FromMultipart",
    "FromFormField",
    "FromFormData",
    "FromHeader",
    "FromJson",
    "FormField",
    "FromPath",
    "FromQuery",
    "HTTPException",
    "FromFile",
    "status",
    "Request",
    "Response",
    "WebSocketRoute",
    "WebSocket",
)



import collections.abc
import enum
import inspect
import typing

from pydantic.fields import ModelField
from starlette.datastructures import UploadFile
from starlette.requests import HTTPConnection, Request

from xpresso._utils.pydantic_utils import model_field_from_param
from xpresso._utils.typing import Literal
from xpresso.binders._binders.media_type_validator import MediaTypeValidator
from xpresso.binders._binders.pydantic_validators import validate_body_field
from xpresso.binders.api import (
    ModelNameMap,
    OpenAPIMetadata,
    SupportsExtractor,
    SupportsOpenAPI,
)
from xpresso.openapi import models as openapi_models
from xpresso.openapi._utils import parse_examples


class FileType(enum.Enum):
    bytes = enum.auto()
    uploadfile = enum.auto()
    stream = enum.auto()


STREAM_TYPES = (typing.AsyncIterator, typing.AsyncGenerator, typing.AsyncIterable, collections.abc.AsyncGenerator, collections.abc.AsyncIterable, collections.abc.AsyncIterator)  # type: ignore


def get_file_type(field: ModelField) -> FileType:
    if field.type_ is bytes:
        return FileType.bytes
    if inspect.isclass(field.type_) and issubclass(field.type_, UploadFile):
        return FileType.uploadfile
    if field.type_ in STREAM_TYPES:  # type: ignore
        return FileType.stream
    raise TypeError(f"Target type {field.type_.__name__} is not recognized")


async def consume_into_bytes(request: Request) -> bytes:
    res = bytearray()
    async for chunk in request.stream():
        res.extend(chunk)
    return res


async def read_into_bytes(request: Request) -> bytes:
    return await request.body()


def create_consume_into_uploadfile(
    cls: typing.Type[UploadFile],
) -> typing.Callable[[Request], typing.Awaitable[UploadFile]]:
    async def consume_into_uploadfile(request: Request) -> UploadFile:
        file = cls(
            filename="body", content_type=request.headers.get("Content-Type", "*/*")
        )
        async for chunk in request.stream():
            if chunk:
                await file.write(chunk)
        await file.seek(0)
        return file

    return consume_into_uploadfile


def create_read_into_uploadfile(
    cls: typing.Type[UploadFile],
) -> typing.Callable[[Request], typing.Awaitable[UploadFile]]:
    async def read_into_uploadfile(request: Request) -> UploadFile:
        file = cls(
            filename="body", content_type=request.headers.get("Content-Type", "*/*")
        )
        await file.write(await request.body())
        await file.seek(0)
        return file

    return read_into_uploadfile


async def consume_into_stream(request: Request) -> typing.AsyncIterator[bytes]:
    return request.stream()


class Extractor(typing.NamedTuple):
    media_type_validator: MediaTypeValidator
    consumer: typing.Callable[[Request], typing.Awaitable[typing.Any]]
    field: ModelField

    def __hash__(self) -> int:
        return hash("file")

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, Extractor)

    async def extract(self, connection: HTTPConnection) -> typing.Any:
        assert isinstance(connection, Request)
        media_type = connection.headers.get("content-type", None)
        if media_type is None and connection.headers.get("content-length", "0") == "0":
            return validate_body_field(None, field=self.field, loc=("body",))
        self.media_type_validator.validate(media_type)
        return await self.consumer(connection)


class ExtractorMarker(typing.NamedTuple):
    media_type: typing.Optional[str]
    enforce_media_type: bool
    consume: bool

    def register_parameter(self, param: inspect.Parameter) -> SupportsExtractor:
        if self.media_type and self.enforce_media_type:
            media_type_validator = MediaTypeValidator(self.media_type)
        else:
            media_type_validator = MediaTypeValidator(None)
        consumer: typing.Callable[[Request], typing.Any]
        field = model_field_from_param(param, arbitrary_types_allowed=True)
        file_type = get_file_type(field)
        if file_type is FileType.bytes:
            if self.consume:
                consumer = consume_into_bytes
            else:
                consumer = read_into_bytes
        elif file_type is FileType.uploadfile:
            if self.consume:
                consumer = create_consume_into_uploadfile(field.type_)
            else:
                consumer = create_read_into_uploadfile(field.type_)
        else:  # stream
            if self.consume:
                consumer = consume_into_stream
            else:
                raise ValueError("consume=False is not supported for streams")
        return Extractor(
            media_type_validator=media_type_validator,
            consumer=consumer,
            field=field,
        )


class OpenAPI(typing.NamedTuple):
    media_type: str
    description: typing.Optional[str]
    examples: typing.Optional[openapi_models.Examples]
    format: Literal["binary", "base64"]
    required: bool
    nullable: bool
    include_in_schema: bool

    def get_models(self) -> typing.List[type]:
        return []

    def get_openapi(
        self,
        model_name_map: ModelNameMap,
    ) -> OpenAPIMetadata:
        if not self.include_in_schema:
            return OpenAPIMetadata()
        return OpenAPIMetadata(
            body=openapi_models.RequestBody(
                content={
                    self.media_type: openapi_models.MediaType(
                        schema=openapi_models.Schema(  # type: ignore
                            type="string",
                            format=self.format,
                            nullable=self.nullable or None,
                        ),
                        examples=self.examples,
                    )
                },
                description=self.description,
                required=self.required,
            )
        )


class OpenAPIMarker(typing.NamedTuple):
    media_type: typing.Optional[str]
    description: typing.Optional[str]
    examples: typing.Optional[
        typing.Dict[str, typing.Union[openapi_models.Example, typing.Any]]
    ]
    format: Literal["binary", "base64"]
    include_in_schema: bool

    def register_parameter(self, param: inspect.Parameter) -> SupportsOpenAPI:
        field = model_field_from_param(param, arbitrary_types_allowed=True)
        examples = parse_examples(self.examples) if self.examples else None
        required = field.required is not False
        return OpenAPI(
            media_type=self.media_type or "*/*",
            description=self.description,
            examples=examples,
            format=self.format,
            required=required,
            nullable=field.allow_none,
            include_in_schema=self.include_in_schema,
        )



from typing import Any, Mapping, Union

from xpresso.encoders import JsonableEncoder
from xpresso.openapi import models as openapi_models
from xpresso.responses import ResponseSpec

ENCODER = JsonableEncoder()


def merge_response_specs(r1: ResponseSpec, r2: ResponseSpec) -> ResponseSpec:
    return ResponseSpec(
        description=r2.description or r1.description,
        headers={**(r2.headers or {}), **(r1.headers or {})} or None,
        content={**(r2.content or {}), **(r1.content or {})} or None,
    )


def parse_examples(
    examples: Mapping[str, Union[openapi_models.Example, Any]]
) -> openapi_models.Examples:
    return {
        k: v
        if isinstance(v, openapi_models.Example)
        else openapi_models.Example(value=ENCODER(v))
        for k, v in examples.items()
    }



import inspect
from http import HTTPStatus
from typing import Any, Dict, Iterable, List, Mapping, Optional, Set, Tuple, Union

from pydantic import BaseConfig
from pydantic.fields import ModelField
from pydantic.schema import field_schema, get_flat_models_from_fields
from pydantic.schema import get_model_name_map as get_model_name_map_pydantic
from starlette.responses import Response
from starlette.routing import compile_path

from xpresso._utils.routing import VisitedRoute
from xpresso._utils.typing import get_args, get_origin, get_type_hints
from xpresso.binders import dependants as binder_dependants
from xpresso.openapi import models
from xpresso.openapi._constants import REF_PREFIX
from xpresso.openapi._utils import merge_response_specs, parse_examples
from xpresso.responses import ResponseModel, ResponseSpec, TypeUnset
from xpresso.routing.operation import Operation
from xpresso.routing.pathitem import Path
from xpresso.routing.router import Router

ModelNameMap = Dict[type, str]

Routes = Mapping[str, Tuple[Path, Mapping[str, Operation]]]


validation_error_definition = {
    "title": "ValidationError",
    "type": "object",
    "properties": {
        "loc": {
            "title": "Location",
            "type": "array",
            "items": {"oneOf": [{"type": "string"}, {"type": "integer"}]},
        },
        "msg": {"title": "Message", "type": "string"},
        "type": {"title": "Error Type", "type": "string"},
    },
    "required": ["loc", "msg", "type"],
}

validation_error_response_definition = {
    "title": "HTTPValidationError",
    "type": "object",
    "properties": {
        "detail": {
            "title": "Detail",
            "type": "array",
            "items": {"$ref": f"{REF_PREFIX}ValidationError"},
        }
    },
}

validation_error_response = models.Response(
    description="Validation Error",
    content={
        "application/json": models.MediaType(
            schema=models.Schema.parse_obj({"$ref": f"{REF_PREFIX}HTTPValidationError"})  # type: ignore
        )
    },
)

status_code_range_descriptions = {
    "1XX": "Information",
    "2XX": "Success",
    "3XX": "Redirection",
    "4XX": "Client Error",
    "5XX": "Server Error",
    "DEFAULT": "Default Response",
}

status_code_descriptions = {
    str(v.value): v.phrase for v in HTTPStatus.__members__.values()
}


def get_model_name_map(unique_models: Set[type]) -> Dict[type, str]:
    # this works with any class, but Pydantic types it as if it only works with Pydantic models
    # if this at some point breaks, we'll just implement it in this function
    return get_model_name_map_pydantic({model for model in unique_models if hasattr(model, "__name__")})  # type: ignore[arg-type]


def get_schema(
    type_: type, model_name_map: ModelNameMap, schemas: Dict[str, Any]
) -> models.Schema:
    field = ModelField.infer(
        name="Response",
        value=...,
        annotation=type_,
        class_validators=None,
        config=BaseConfig,
    )
    flat_models = get_flat_models_from_fields([field], known_models=set())
    model_name_map = get_model_name_map(flat_models)
    schema, new_schemas, _ = field_schema(field, model_name_map=model_name_map, ref_prefix=REF_PREFIX)  # type: ignore[arg-type]
    if "title" in schema and schema["title"] == "Response":
        schema.pop("title", None)
    schemas.update(new_schemas)
    return models.Schema(**schema)


def description_from_user_input_or_status_code(
    description: Optional[str], status_code: str
) -> str:
    if description:
        return description
    if status_code in status_code_descriptions:
        return status_code_descriptions[status_code]
    if status_code in status_code_range_descriptions:
        return status_code_range_descriptions[status_code]
    raise ValueError(f'Unknown status code "{status_code}"')


def get_response_model(
    spec: ResponseSpec,
    status_code: str,
    model_name_map: ModelNameMap,
    schemas: Dict[str, Any],
) -> models.Response:
    headers = {
        header: models.ResponseHeader(description=header_or_description)
        if isinstance(header_or_description, str)
        else header_or_description
        for header, header_or_description in (spec.headers or {}).items()
    } or None
    content = {
        k: v if isinstance(v, ResponseModel) else ResponseModel(v)
        for k, v in (spec.content or {}).items()
    }
    examples = {
        k: parse_examples(v.examples) if v.examples is not None else None
        for k, v in content.items()
    }
    schemas = {
        k: get_schema(v.model, model_name_map, schemas)
        if v.model is not TypeUnset
        else None
        for k, v in content.items()
    }
    return models.Response(
        description=description_from_user_input_or_status_code(
            spec.description, status_code
        ),
        headers=headers,  # type: ignore[arg-type]
        content={
            k: models.MediaType(
                schema=schemas[k],  # type: ignore
                examples=examples[k],
            )
            for k in content
        }
        or None,
    )


def get_responses(
    response_specs: Mapping[str, ResponseSpec],
    model_name_map: ModelNameMap,
    schemas: Dict[str, Any],
) -> Dict[str, models.Response]:
    responses: Dict[str, models.Response] = {}
    for status, response_spec in response_specs.items():
        if (
            status in responses
            or f"{status[0]}XX" in responses
            or (
                status.endswith("XX")
                and any(str(s).startswith(status[0]) for s in responses)
            )
        ):
            raise ValueError("Duplicate response status codes are not allowed")
        responses[status] = get_response_model(
            response_spec, status, model_name_map, schemas
        )
    return responses


def is_response(tp: type) -> bool:
    return inspect.isclass(tp) and issubclass(tp, Response)


def get_operation(
    route: Operation,
    model_name_map: ModelNameMap,
    components: Dict[str, Any],
    tags: List[str],
    response_specs: Dict[str, ResponseSpec],
) -> models.Operation:
    data: Dict[str, Any] = {
        "tags": tags or None,
        "summary": route.summary,
        "description": route.description,
        "deprecated": route.deprecated,
        "servers": route.servers or None,
        "externalDocs": route.external_docs,
        "operationId": route.operation_id,
    }
    docstring = getattr(route.endpoint, "__doc__", None)
    if docstring and not data["description"]:
        data["description"] = docstring
    route_dependant = route.dependant
    openapis = [
        dep.openapi.get_openapi(model_name_map)
        for dep in route_dependant.get_flat_subdependants()
        if isinstance(dep, binder_dependants.Binder)
    ]
    schemas = {
        name: schema for oai in openapis for name, schema in (oai.schemas or {}).items()
    }
    params = [param for oai in openapis for param in oai.parameters or ()]
    request_body: Optional[models.RequestBody] = None
    for oai in openapis:
        if oai.body:
            if request_body:
                raise ValueError("Only 1 top level body is allowed")
            # sort so that generated openapi is deterministic
            request_body_content = dict(
                sorted(
                    [(k, v) for k, v in oai.body.content.items()], key=lambda kv: kv[0]
                )
            )
            request_body = models.RequestBody(
                content=request_body_content,
                required=oai.body.required,
                description=oai.body.description,
            )
    # sort so that generated openapi is deterministic
    params = list(sorted(params, key=lambda param: param.name))
    data["parameters"] = params or None
    data["requestBody"] = request_body
    # merge in the default response spec
    response_model = route.response_model
    if response_model is TypeUnset:
        sig_return = inspect.signature(route.endpoint).return_annotation
        if sig_return is not inspect.Parameter.empty:
            response_annotation = get_type_hints(route.endpoint)["return"]
            if (
                # get_type_hints returns type(None)
                # if the func is () -> None we don't add a response model
                # it is rare to want to _document_ "null" as the response model
                sig_return
                is None
            ) or (
                # this is a special case for () -> FileResponse and the like
                is_response(response_annotation)
                or get_origin(response_annotation) is Union
                and any(is_response(tp) for tp in get_args(response_annotation))
            ):
                response_annotation = TypeUnset
            if response_annotation is not TypeUnset:
                response_model = response_annotation
    default_content = {
        route.response_media_type: ResponseModel(
            model=response_model,
            examples=route.response_examples,
        )
    }
    route_response_status_code = str(route.response_status_code)
    route_response_description = description_from_user_input_or_status_code(
        route.response_description, route_response_status_code
    )
    if route_response_status_code in response_specs:
        if response_specs[route_response_status_code].content:
            content = response_specs[route_response_status_code].content
        else:
            content = default_content
        response_specs[route_response_status_code] = merge_response_specs(
            ResponseSpec(
                description=route_response_description,
                content=content,
                headers=route.response_headers,
            ),
            response_specs[route_response_status_code],
        )
    else:
        response_specs[route_response_status_code] = ResponseSpec(
            description=route_response_description,
            content=default_content,
            headers=route.response_headers,
        )
    data["responses"] = get_responses(
        response_specs=response_specs,
        model_name_map=model_name_map,
        schemas=schemas,
    )
    if schemas:
        components["schemas"] = {**components.get("schemas", {}), **schemas}
    if ((data.get("parameters", None) or data.get("requestBody", None))) and all(
        status not in data["responses"] for status in ("422", "4XX", "default")
    ):
        data["responses"]["422"] = validation_error_response

        if "ValidationError" not in schemas:
            components["schemas"] = components.get("schemas", None) or {}
            components["schemas"].update(
                {
                    "ValidationError": validation_error_definition,
                    "HTTPValidationError": validation_error_response_definition,
                }
            )
    return models.Operation(**data)


def merge_node_openapi_metadata(
    node: Union[Router, Path, Operation],
    tags: List[str],
    responses: Dict[str, ResponseSpec],
) -> Tuple[List[str], Dict[str, ResponseSpec]]:
    new_responses: Dict[str, ResponseSpec] = responses.copy()
    for status_code, response in node.responses.items():
        status_code_str = str(status_code)
        if status_code not in responses:
            new_responses[status_code_str] = response
        else:
            new_responses[status_code_str] = merge_response_specs(
                responses[status_code_str], response
            )
    return [*tags, *node.tags], new_responses


def get_paths_items(
    visitor: Iterable[VisitedRoute[Any]],
    model_name_map: ModelNameMap,
    components: Dict[str, Any],
) -> Dict[str, models.PathItem]:
    paths: "Dict[str, models.PathItem]" = {}
    for visited_route in visitor:
        if isinstance(visited_route.route, Path):
            path_item = visited_route.route
            if not path_item.include_in_schema:
                continue
            tags: "List[str]" = []
            include_in_schema = True
            responses: "Dict[str, ResponseSpec]" = {}
            for node in visited_route.nodes:
                if isinstance(node, Router):
                    if not node.include_in_schema:
                        include_in_schema = False
                        break
                    tags, responses = merge_node_openapi_metadata(node, tags, responses)
            if not include_in_schema:
                continue
            tags, responses = merge_node_openapi_metadata(path_item, tags, responses)
            operations: "Dict[str, models.Operation]" = {}
            for method, operation in path_item.operations.items():
                if not operation.include_in_schema:
                    continue
                operation_tags, operation_responses = merge_node_openapi_metadata(
                    operation, tags, responses
                )
                operations[method.lower()] = get_operation(
                    operation,
                    model_name_map=model_name_map,
                    components=components,
                    tags=operation_tags,
                    response_specs=operation_responses,
                )
            path = compile_path(visited_route.path)[1]
            paths[path] = models.PathItem(
                description=visited_route.route.description,
                summary=visited_route.route.summary,
                servers=list(visited_route.route.servers) or None,
                **operations,  # type: ignore[arg-type]
            )  # type: ignore  # for Pylance
    return {k: paths[k] for k in sorted(paths.keys())}


def filter_routes(visitor: Iterable[VisitedRoute[Any]]) -> Routes:
    res: Dict[str, Tuple[Path, Dict[str, Operation]]] = {}
    for visited_route in visitor:
        if isinstance(visited_route.route, Path):
            path_item = visited_route.route
            if not path_item.include_in_schema:
                continue
            operations: Dict[str, Operation] = {
                method.lower(): operation
                for method, operation in path_item.operations.items()
                if operation.include_in_schema
            }

            res[visited_route.path] = (path_item, operations)
    return res


def get_flat_models(routes: Routes) -> Set[type]:
    res: Set[type] = set()
    for _, operations in routes.values():
        for operation in operations.values():
            dependant = operation.dependant
            flat_dependencies = dependant.get_flat_subdependants()
            for dep in flat_dependencies:
                if isinstance(
                    dep,
                    binder_dependants.Binder,
                ):
                    res.update(dep.openapi.get_models())
            for response in operation.responses.values():
                for response_model in (response.content or {}).values():
                    if (
                        isinstance(response_model, ResponseModel)
                        and response_model.model is not TypeUnset
                    ):
                        res.add(response_model.model)
    return res


def generate_openapi(
    visitor: Iterable[VisitedRoute[Any]],
    version: str,
    info: models.Info,
    servers: Optional[Iterable[models.Server]],
) -> models.OpenAPI:
    visitor = list(visitor)
    routes = filter_routes(visitor)
    flat_models = get_flat_models(routes)
    model_name_map = get_model_name_map(flat_models)
    components: Dict[str, Any] = {}
    paths = get_paths_items(visitor, model_name_map, components)
    return models.OpenAPI(
        openapi=version,
        info=info,
        paths=paths,  # type: ignore[arg-type]
        components=models.Components(**components) if components else None,
        servers=list(servers) if servers else None,
    )
